"""Define a functions for modifying incoming data."""
from datetime import datetime, timedelta
from typing import Any, Optional, Union

import pytz

from .exceptions import BlueCurrentException, RequestLimitReached, WebsocketError

TZ = pytz.timezone("Europe/Amsterdam")

ERRORS: dict[int, BlueCurrentException] = {
    0: WebsocketError("Unknown command"),
    1: WebsocketError("Invalid Auth Token"),
    2: WebsocketError("Not authorized"),
    9: WebsocketError("Unknown error"),
    42: RequestLimitReached("Request limit reached"),
}

SMART_CHARGING: set[str] = set()


def calculate_average_usage_from_phases(phases: list[float | None]) -> float:
    """Get the average of the phases that are not 0."""
    used_phases = [p for p in phases if p]
    if len(used_phases):
        return round(sum(used_phases) / len(used_phases), 1)
    return 0


def calculate_total_kw(c_avg: float, v_avg: float) -> float:
    """Calculate the total kW."""
    return round((c_avg * v_avg * 1.732 / 1000), 2)


def create_datetime(timestamp: str) -> Optional[datetime]:
    """Get a datetime object from an timestamp."""

    if timestamp == "":
        return None

    if "+" in timestamp:
        return datetime.strptime(timestamp, "%Y%m%d %H:%M:%S%z")

    time = datetime.strptime(timestamp, "%Y%m%d %H:%M:%S")
    time = TZ.localize(time)
    return time


def get_vehicle_status(vehicle_status_key: str) -> str:
    """Get the vehicle status."""
    statuses = {
        "A": "standby",
        "B": "vehicle_detected",
        "C": "ready",
        "D": "ready",
        "E": "no_power",
        "F": "vehicle_error",
    }

    return statuses[vehicle_status_key]


def get_exception(message: dict[str, Any]) -> BlueCurrentException:
    """Return a defined error message or one from the server"""
    error = message["error"]
    message = message["message"]

    if error in ERRORS:
        return ERRORS[error]
    return WebsocketError(message)


def set_smart_charging(evse_id: str, smart_charging: bool) -> None:
    """Add or discard evse_id in SMART_CHARGING"""
    if smart_charging:
        SMART_CHARGING.add(evse_id)
    else:
        SMART_CHARGING.discard(evse_id)


def handle_charge_points(message: dict[str, Any]) -> None:
    """Store the evse_id if it has smart charging enabled"""
    for charge_point in message["data"]:
        set_smart_charging(charge_point["evse_id"], charge_point["smart_charging"])


def set_current_left(message: dict[str, Any], c_avg: float) -> None:
    """Set current_left"""
    max_usage = message["data"]["max_usage"]
    smart_charging_max_usage = message["data"]["smartcharging_max_usage"]

    if message["evse_id"] in SMART_CHARGING:
        message["data"]["current_left"] = smart_charging_max_usage - c_avg
    else:
        message["data"]["current_left"] = max_usage - c_avg


def handle_status(message: dict[str, Any]) -> None:
    """Transform status values and add others."""
    voltage1: float = message["data"]["actual_v1"]
    voltage2: float = message["data"]["actual_v2"]
    voltage3: float = message["data"]["actual_v3"]

    v_avg = calculate_average_usage_from_phases([voltage1, voltage2, voltage3])
    message["data"]["avg_voltage"] = v_avg

    current1: float = message["data"]["actual_p1"]
    current2: float = message["data"]["actual_p2"]
    current3: float = message["data"]["actual_p3"]

    c_avg = calculate_average_usage_from_phases([current1, current2, current3])
    message["data"]["avg_current"] = c_avg

    set_current_left(message, c_avg)

    message["data"]["total_kw"] = calculate_total_kw(c_avg, v_avg)

    vehicle_status_key = message["data"]["vehicle_status"]
    message["data"]["vehicle_status"] = get_vehicle_status(vehicle_status_key)

    start_datetime = message["data"]["start_datetime"]
    new_start_datetime = create_datetime(start_datetime)

    stop_datetime = message["data"]["stop_datetime"]
    new_stop_datetime = create_datetime(stop_datetime)

    message["data"]["start_datetime"] = new_start_datetime
    message["data"]["stop_datetime"] = new_stop_datetime

    offline_since = message["data"]["offline_since"]
    message["data"]["offline_since"] = create_datetime(offline_since)


def handle_settings(message: dict[str, Any]) -> None:
    """Transform settings object"""

    message["data"]["plug_and_charge"] = message["data"]["plug_and_charge"]["value"]
    message["data"]["linked_charge_cards_only"] = not message["data"][
        "public_charging"
    ]["value"]

    set_smart_charging(message["data"]["evse_id"], message["data"]["smart_charging"])


def handle_grid(message: dict[str, Any]) -> None:
    """Add grid total and avg to a message."""
    if "CURRENT" in message["object"]:
        message["data"] = {}
        message["data"]["grid_actual_p1"] = message.pop("grid_actual_p1")
        message["data"]["grid_actual_p2"] = message.pop("grid_actual_p2")
        message["data"]["grid_actual_p3"] = message.pop("grid_actual_p3")

    current1: float = message["data"]["grid_actual_p1"]
    current2: float = message["data"]["grid_actual_p2"]
    current3: float = message["data"]["grid_actual_p3"]

    c_avg = calculate_average_usage_from_phases([current1, current2, current3])
    message["data"]["grid_avg_current"] = c_avg
    c_max = max(current1, current2, current3)
    message["data"]["grid_max_current"] = c_max


def handle_setting_change(message: dict[str, Any]) -> None:
    """Change result to a boolean."""
    message["result"] = "true" in message["result"]["setting"]

    if message["object"] == "STATUS_SET_PUBLIC_CHARGING":
        message["object"] = "LINKED_CHARGE_CARDS_ONLY"
        message["result"] = not message["result"]
    else:
        message["object"] = message["object"].replace("STATUS_SET_", "")


def handle_session_messages(message: dict[str, Any]) -> None:
    """handle session messages."""

    object_name = message["object"].replace("STATUS_", "").replace("RECEIVED_", "")

    if "STATUS" in message["object"] and message["error"]:
        name = object_name.lower()
        error = message["error"].lower()
        evse_id = message["evse_id"]
        message["error"] = f"{name} {error} for chargepoint: {evse_id}"
    message["object"] = object_name


def get_dummy_message(evse_id: str) -> dict[str, Union[str, dict[str, Any]]]:
    """Return a CH_STATUS message with the current time as start_datetime"""
    return {
        "object": "CH_STATUS",
        "data": {
            "start_datetime": datetime.now(TZ),
            "evse_id": evse_id,
        },
    }


def get_next_reset_delta() -> timedelta:
    """Returns the timedelta to the next midnight"""
    now = datetime.now(TZ)
    return now.replace(hour=0, minute=0, second=30) + timedelta(days=1) - now



from unittest import mock
from src.bluecurrent_api.utils import (
    calculate_average_usage_from_phases,
    get_vehicle_status,
    create_datetime,
    handle_charge_points,
    handle_grid,
    handle_session_messages,
    handle_settings,
    handle_status,
    get_dummy_message,
    get_exception,
    get_next_reset_delta,
    calculate_total_kw,
    SMART_CHARGING,
    set_current_left,
    set_smart_charging,
    handle_setting_change,
)
from datetime import datetime, timezone, timedelta
from src.bluecurrent_api.exceptions import WebsocketError, RequestLimitReached
from pytest_mock import MockerFixture
import pytz


def test_calculate_total_from_phases():
    total = calculate_average_usage_from_phases([10, 10, 10])
    assert total == 10

    total = calculate_average_usage_from_phases([0, 0, 0])
    assert total == 0

    total = calculate_average_usage_from_phases([10, None, 20])
    assert total == 15

    total = calculate_average_usage_from_phases([10, 6, 10])
    assert total == 8.7

    total = calculate_average_usage_from_phases([10, 8, 1])
    assert total == 6.3

    total = calculate_average_usage_from_phases([10, 0, 20])
    assert total == 15

    total = calculate_average_usage_from_phases([5, 0, 0])
    assert total == 5

    total = calculate_average_usage_from_phases([6, None, None])
    assert total == 6

    total = calculate_average_usage_from_phases([None, None, None])
    assert total == 0


def test_calculate_total_kW():
    total = calculate_total_kw(14, 230)
    assert total == 5.58


def test_set_to_smart_charging():
    set_smart_charging("bcu101", True)
    assert SMART_CHARGING == {"bcu101"}
    set_smart_charging("bcu101", False)
    assert SMART_CHARGING == set()


def test_get_vehicle_status():
    assert get_vehicle_status("A") == "standby"
    assert get_vehicle_status("B") == "vehicle_detected"
    assert get_vehicle_status("C") == "ready"
    assert get_vehicle_status("D") == "ready"
    assert get_vehicle_status("E") == "no_power"
    assert get_vehicle_status("F") == "vehicle_error"


def test_create_datetime():
    TZ = pytz.timezone("Europe/Amsterdam")

    test_time = TZ.localize(datetime(2001, 1, 1, 0, 0, 0))

    assert create_datetime("20010101 00:00:00") == test_time

    assert create_datetime("20010101 00:00:00+00:00") == datetime(
        2001, 1, 1, 0, 0, 0, tzinfo=timezone.utc
    )

    assert create_datetime("") is None


def test_handle_charge_points():
    message = {
        "data": [
            {"evse_id": "BCU101", "smart_charging": False},
            {"evse_id": "BCU102", "smart_charging": True},
        ]
    }

    handle_charge_points(message)
    assert SMART_CHARGING == {"BCU102"}


def test_handle_set_current_left():
    message = {
        "data": {
            "max_usage": 10,
            "smartcharging_max_usage": 6,
        },
        "evse_id": "bcu101",
    }

    SMART_CHARGING.add("bcu101")
    set_current_left(message, 5)
    assert message["data"]["current_left"] == 1

    SMART_CHARGING.remove("bcu101")
    set_current_left(message, 5)
    assert message["data"]["current_left"] == 5


def test_handle_status():
    message = {
        "data": {
            "actual_v1": 220,
            "actual_v2": 221,
            "actual_v3": 219,
            "actual_p1": 8,
            "actual_p2": 8,
            "actual_p3": 8,
            "activity": "available",
            "start_datetime": "20211118 14:12:23",
            "stop_datetime": "20211118 14:32:23",
            "offline_since": "20211118 14:32:23",
            "total_cost": 10.52,
            "vehicle_status": "A",
            "actual_kwh": 10,
            "evse_id": "BCU101",
            "max_usage": 10,
            "max_offline": 7,
            "smartcharging_max_usage": 6,
        },
        "evse_id": "BCU101",
    }

    TZ = pytz.timezone("Europe/Amsterdam")

    handle_status(message)

    assert message["data"]["avg_voltage"] == 220.0
    assert message["data"]["avg_current"] == 8.0
    assert message["data"]["total_kw"] == 3.05
    assert message["data"]["start_datetime"] == TZ.localize(
        datetime(2021, 11, 18, 14, 12, 23)
    )
    assert message["data"]["stop_datetime"] == TZ.localize(
        datetime(2021, 11, 18, 14, 32, 23)
    )
    assert message["data"]["offline_since"] == TZ.localize(
        datetime(2021, 11, 18, 14, 32, 23)
    )
    assert message["data"]["vehicle_status"] == "standby"
    assert message["data"]["current_left"] == 2.0

    assert len(message["data"]) == 21


def test_handle_settings():
    message = {
        "data": {
            "evse_id": "BCU102",
            "plug_and_charge": {
                "value": False,
            },
            "public_charging": {
                "value": True,
            },
            "smart_charging": False,
        }
    }
    handle_settings(message)
    assert message["data"]["plug_and_charge"] is False
    assert message["data"]["linked_charge_cards_only"] is False
    assert SMART_CHARGING == set()

    message = {
        "data": {
            "evse_id": "BCU102",
            "plug_and_charge": {
                "value": False,
            },
            "public_charging": {
                "value": True,
            },
            "smart_charging": True,
        }
    }
    handle_settings(message)
    assert SMART_CHARGING == {"BCU102"}


def test_get_exception():
    message = {"object": "ERROR", "error": 0, "message": "Test error"}

    error = get_exception(message)
    assert isinstance(error, WebsocketError)
    assert str(error) == "Unknown command"

    message = {"object": "ERROR", "error": 99, "message": "Test error"}
    error = get_exception(message)
    assert isinstance(error, WebsocketError)
    assert str(error) == "Test error"

    message = {"object": "ERROR", "error": 42, "message": "Test error"}
    error = get_exception(message)
    assert isinstance(error, RequestLimitReached)
    assert str(error) == "Request limit reached"


def test_handle_grid():
    message = {
        "data": {
            "grid_actual_p1": 12,
            "grid_actual_p2": 14,
            "grid_actual_p3": 15,
        },
        "object": "GRID_STATUS",
    }

    handle_grid(message)
    assert message["data"]["grid_avg_current"] == 13.7
    assert message["data"]["grid_max_current"] == 15

    message = {
        "grid_actual_p1": 12,
        "grid_actual_p2": 14,
        "grid_actual_p3": 15,
        "object": "GRID_CURRENT",
    }

    handle_grid(message)
    assert message["data"]["grid_avg_current"] == 13.7
    assert message["data"]["grid_max_current"] == 15

    assert len(message["data"]) == 5


def test_handle_setting_change():
    message = {
        "object": "STATUS_SET_PLUG_AND_CHARGE",
        "result": {"setting": "set true"},
    }

    handle_setting_change(message)
    assert message == {"object": "PLUG_AND_CHARGE", "result": True}

    message = {
        "object": "STATUS_SET_PUBLIC_CHARGING",
        "result": {"setting": "set false"},
    }

    handle_setting_change(message)
    assert message == {"object": "LINKED_CHARGE_CARDS_ONLY", "result": True}


def test_handle_session_messages():
    message = {
        "object": "RECEIVED_STOP_SESSION",
        "success": False,
        "error": "no active session for chargepoint: BCU101",
    }
    handle_session_messages(message)
    assert message == {
        "object": "STOP_SESSION",
        "success": False,
        "error": "no active session for chargepoint: BCU101",
    }

    message = {
        "object": "STATUS_START_SESSION",
        "success": False,
        "evse_id": "BCU101",
        "error": "REJECTED",
    }
    handle_session_messages(message)
    assert message == {
        "object": "START_SESSION",
        "success": False,
        "evse_id": "BCU101",
        "error": "start_session rejected for chargepoint: BCU101",
    }

    message = {
        "object": "STATUS_START_SESSION",
        "success": False,
        "evse_id": "BCU101",
        "error": "TIMEOUT",
    }
    handle_session_messages(message)
    assert message == {
        "object": "START_SESSION",
        "success": False,
        "evse_id": "BCU101",
        "error": "start_session timeout for chargepoint: BCU101",
    }

    message = {
        "object": "STATUS_REBOOT",
        "success": False,
        "evse_id": "BCU101",
        "error": "TIMEOUT",
    }
    handle_session_messages(message)
    assert message == {
        "object": "REBOOT",
        "success": False,
        "evse_id": "BCU101",
        "error": "reboot timeout for chargepoint: BCU101",
    }

    message = {
        "object": "STATUS_SOFT_RESET",
        "success": False,
        "evse_id": "BCU101",
        "error": "TIMEOUT",
    }
    handle_session_messages(message)
    assert message == {
        "object": "SOFT_RESET",
        "success": False,
        "evse_id": "BCU101",
        "error": "soft_reset timeout for chargepoint: BCU101",
    }


def test_get_dummy_message(mocker: MockerFixture):
    time = datetime(1901, 12, 21)
    datetime_mock = mocker.patch("src.bluecurrent_api.utils.datetime")
    datetime_mock.now = mock.Mock(return_value=time)
    assert get_dummy_message("BCU101") == {
        "object": "CH_STATUS",
        "data": {"start_datetime": time, "evse_id": "BCU101"},
    }


def test_get_next_reset_delta(mocker: MockerFixture):
    time = datetime(1901, 12, 21, 10, 30, 0)
    datetime_mock = mocker.patch("src.bluecurrent_api.utils.datetime")
    datetime_mock.now = mock.Mock(return_value=time)
    assert get_next_reset_delta() == timedelta(hours=13, minutes=30, seconds=30)



import os
import sys
sys.path.insert(0, os.path.abspath('../../'))


# -- Project information -----------------------------------------------------

project = 'Blue Current HA integration Docs'
copyright = '2022, Floris Puijk'
author = 'Floris Puijk'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'myst_parser',
    'matplotlib.sphinxext.plot_directive',
    'sphinx.ext.autodoc'
]

# Allow both restructuredText and MarkDown syntax
source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'markdown',
    '.md': 'markdown',
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

html_theme_options = {
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    'vcs_pageview_mode': '',
    # Toc options
    'collapse_navigation': False,
    'sticky_navigation': True,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': False
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add static js files to allow for offline building :)
html_js_files = [
    'js/d3.v6.min.js'
]

html_css_files = [
    'css/layout.css',
]
# Suppress specific warnings
suppress_warnings = ["myst.header", '']
myst_heading_anchors = 3

autodoc_member_order = 'bysource'



from src.bluecurrent_api.client import Client
import pytest
from pytest_mock import MockerFixture


def test_create_request():
    client = Client()

    # command
    request = client._create_request("GET_CHARGE_POINTS")
    assert request == {"command": "GET_CHARGE_POINTS"}

    # evse_id
    request = client._create_request("GET_STATUS", "101")
    assert request == {"command": "GET_STATUS", "evse_id": "101"}

    # value
    request = client._create_request("SET_PLUG_AND_CHARGE", "101", True)
    assert request == {
        "command": "SET_PLUG_AND_CHARGE",
        "evse_id": "101",
        "value": True,
    }

    # card_uid / session_token
    request = client._create_request("START_SESSION", "101", card_uid="1234")
    assert request == {
        "command": "START_SESSION",
        "evse_id": "101",
        "session_token": "1234",
    }


@pytest.mark.asyncio
async def test_requests(mocker: MockerFixture):
    test_send_request = mocker.patch(
        "src.bluecurrent_api.client.Websocket.send_request"
    )
    client = Client()

    await client.get_charge_cards()
    test_send_request.assert_called_with({"command": "GET_CHARGE_CARDS", "limit": 100})

    await client.get_charge_points()
    test_send_request.assert_called_with({"command": "GET_CHARGE_POINTS"})

    await client.get_status("101")
    test_send_request.assert_called_with({"command": "GET_CH_STATUS", "evse_id": "101"})

    await client.get_settings("101")
    test_send_request.assert_called_with(
        {"command": "GET_CH_SETTINGS", "evse_id": "101"}
    )

    await client.get_grid_status("101")
    test_send_request.assert_called_with(
        {"command": "GET_GRID_STATUS", "evse_id": "101"}
    )

    await client.set_linked_charge_cards_only("101", True)
    test_send_request.assert_called_with(
        {"command": "SET_PUBLIC_CHARGING", "evse_id": "101", "value": False}
    )

    await client.set_plug_and_charge("101", True)
    test_send_request.assert_called_with(
        {"command": "SET_PLUG_AND_CHARGE", "evse_id": "101", "value": True}
    )

    await client.block("101", False)
    test_send_request.assert_called_with({"command": "SET_OPERATIVE", "evse_id": "101"})

    await client.block("101", True)
    test_send_request.assert_called_with(
        {"command": "SET_INOPERATIVE", "evse_id": "101"}
    )

    await client.reset("101")
    test_send_request.assert_called_with({"command": "SOFT_RESET", "evse_id": "101"})

    await client.reboot("101")
    test_send_request.assert_called_with({"command": "REBOOT", "evse_id": "101"})

    await client.start_session("101", "123")
    test_send_request.assert_called_with(
        {"command": "START_SESSION", "evse_id": "101", "session_token": "123"}
    )

    await client.stop_session("101")
    test_send_request.assert_called_with({"command": "STOP_SESSION", "evse_id": "101"})


@pytest.mark.asyncio
async def test_on_open(mocker: MockerFixture):
    test_send_request = mocker.patch(
        "src.bluecurrent_api.client.Websocket.send_request"
    )
    client = Client()

    async def on_open():
        await client.get_charge_cards()
        await client.get_charge_points()

    await client._on_open(on_open)
    test_send_request.assert_has_calls(
        [
            mocker.call(
                {
                    "command": "HELLO",
                    "header": "homeassistant",
                }
            ),
            mocker.call({"command": "GET_CHARGE_CARDS", "limit": 100}),
            mocker.call({"command": "GET_CHARGE_POINTS"}),
        ]
    )



"""Define an object to interact with the BlueCurrent websocket api."""
import logging
from datetime import timedelta
from typing import Any, Optional
from collections.abc import Callable, Coroutine

from .utils import get_next_reset_delta
from .websocket import Websocket

LOGGER = logging.getLogger(__package__)
DELAY = 10


class Client:
    """Api Client for the BlueCurrent Websocket Api."""

    def __init__(self) -> None:
        """Initialize the Client."""
        self.websocket = Websocket()

    def is_connected(self) -> bool:
        """Return the connection status"""
        return self.websocket.connected.is_set()

    async def wait_for_charge_points(self) -> None:
        """Wait for next response."""
        await self.websocket.received_charge_points.wait()

    async def validate_api_token(self, api_token: str) -> str:
        """Validate an api_token and return customer id."""
        return await self.websocket.validate_api_token(api_token)

    async def get_email(self) -> str:
        """Get user email."""
        return await self.websocket.get_email()

    async def _on_open(self, on_open: Callable[[], Coroutine[Any, Any, None]],) -> None:
        """Send requests when connected."""
        await self.websocket.send_request(
            {
                "command": "HELLO",
                "header": "homeassistant",
            }
        )
        await on_open()

    def get_next_reset_delta(self) -> timedelta:
        """Returns the timedelta until the websocket limits are reset."""
        return get_next_reset_delta()

    async def connect(
        self,
        receiver: Callable[[dict[str, Any]], Coroutine[Any, Any, None]],
        on_open: Callable[[], Coroutine[Any, Any, None]],
    ) -> None:
        """Connect to the websocket."""
        await self.websocket.start(receiver, lambda : self._on_open(on_open))

    async def disconnect(self) -> None:
        """Disconnect the websocket."""
        await self.websocket.disconnect()

    async def get_charge_cards(self) -> None:
        """Get the charge cards."""
        await self.websocket.send_request({"command": "GET_CHARGE_CARDS", "limit": 100})

    async def get_charge_points(self) -> None:
        """Get the charge points."""
        request = self._create_request("GET_CHARGE_POINTS")
        await self.websocket.send_request(request)

    async def get_status(self, evse_id: str) -> None:
        """Get the status of a charge point."""
        request = self._create_request("GET_CH_STATUS", evse_id)
        await self.websocket.send_request(request)

    async def get_settings(self, evse_id: str) -> None:
        """Get the settings of a charge point."""
        request = self._create_request("GET_CH_SETTINGS", evse_id)
        await self.websocket.send_request(request)

    async def get_grid_status(self, evse_id: str) -> None:
        """Get the grid status of a charge point."""
        request = self._create_request("GET_GRID_STATUS", evse_id)
        await self.websocket.send_request(request)

    async def set_linked_charge_cards_only(self, evse_id: str, value: bool) -> None:
        """Set public_charging of a charge point to a value."""
        request = self._create_request("SET_PUBLIC_CHARGING", evse_id, not value)
        await self.websocket.send_request(request)

    async def set_plug_and_charge(self, evse_id: str, value: bool) -> None:
        """Set plug_and_charge of a charge point to a value."""
        request = self._create_request("SET_PLUG_AND_CHARGE", evse_id, value)
        await self.websocket.send_request(request)

    async def block(self, evse_id: str, value: bool) -> None:
        """Set available of a charge point to a value."""
        command = "SET_OPERATIVE"
        if value is True:
            command = "SET_INOPERATIVE"
        request = self._create_request(command, evse_id)
        await self.websocket.send_request(request)

    async def reset(self, evse_id: str) -> None:
        """Reset a charge point."""
        request = self._create_request("SOFT_RESET", evse_id)
        await self.websocket.send_request(request)

    async def reboot(self, evse_id: str) -> None:
        """Reboot a charge point."""
        request = self._create_request("REBOOT", evse_id)
        await self.websocket.send_request(request)

    async def start_session(self, evse_id: str, card_uid: str) -> None:
        """Start a charge session at a charge point."""
        request = self._create_request("START_SESSION", evse_id, card_uid=card_uid)
        await self.websocket.send_request(request)

    async def stop_session(self, evse_id: str) -> None:
        """Stop a charge session at a charge point."""
        request = self._create_request("STOP_SESSION", evse_id)
        await self.websocket.send_request(request)

    def _create_request(
        self,
        command: str,
        evse_id: Optional[str] = None,
        value: Optional[bool] = None,
        card_uid: Optional[str] = None,
    ) -> dict[str, Any]:
        """Create a request."""
        request: dict[str, Any] = {"command": command}

        if evse_id:
            request["evse_id"] = evse_id

        if value is not None:
            request["value"] = value

        if card_uid:
            request["session_token"] = card_uid

        return request



"""Define an object that handles the connection to the Websocket"""
import json
from asyncio import Event, timeout
import logging
from typing import Any, cast
from collections.abc import Callable, Coroutine

from websockets.client import connect, WebSocketClientProtocol
from websockets.exceptions import (
    ConnectionClosedError,
    InvalidStatusCode,
    WebSocketException,
)

from .exceptions import (
    AlreadyConnected,
    InvalidApiToken,
    RequestLimitReached,
    WebsocketError,
)
from .utils import (
    get_exception,
    handle_charge_points,
    handle_grid,
    handle_session_messages,
    handle_setting_change,
    handle_settings,
    handle_status,
)

URL = "wss://motown.bluecurrent.nl/haserver"
BUTTONS = ("START_SESSION", "STOP_SESSION", "SOFT_RESET", "REBOOT")
LOGGER = logging.getLogger(__package__)


class Websocket:
    """Class for handling requests and responses for the BlueCurrent Websocket Api."""

    def __init__(self) -> None:
        self.conn: WebSocketClientProtocol | None = None
        self.auth_token: str | None = None
        self.connected = Event()
        self.received_charge_points = Event()

    async def start(
        self,
        receiver: Callable[[dict[str, Any]], Coroutine[Any, Any, None]],
        on_open: Callable[[], Coroutine[Any, Any, None]],
    ) -> None:
        """Opens the connection"""

        if not self.auth_token:
            raise WebsocketError("token not validated.")

        try:
            await self._loop(receiver, on_open)
        except WebSocketException as err:
            self.raise_correct_exception(err)

    async def _loop(
        self,
        receiver: Callable[[dict[str, Any]], Coroutine[Any, Any, None]],
        on_open: Callable[[], Coroutine[Any, Any, None]],
    ) -> None:
        """listens for incomming messages"""
        async with connect(URL) as websocket:
            try:
                LOGGER.debug("connected")
                self.conn = websocket
                self.connected.set()
                await on_open()
                async for message in websocket:
                    await self._message_handler(json.loads(message), receiver)
            except WebSocketException as err:
                self.conn = None
                self.connected.clear()
                self.received_charge_points.set()
                self.received_charge_points.clear()
                self.raise_correct_exception(err)

    async def _send_recv_single_message(self, message_object: dict) -> dict:
        """Send and recv single message."""
        message = json.dumps(message_object)
        try:
            async with connect(URL) as websocket:
                await websocket.send(message)
                async with timeout(5):
                    res = await websocket.recv()
                    return cast(dict, json.loads(res))
        except WebSocketException as err:
            self.raise_correct_exception(err)
            # unreachable since raise_correct_exception will always return an error
            # added for type hints.
            return {}

    async def validate_api_token(self, api_token: str) -> str:
        """Validate an api token."""
        res = await self._send_recv_single_message(
            {"command": "VALIDATE_API_TOKEN", "token": api_token}
        )

        if res["object"] == "ERROR":
            raise get_exception(res)

        if not res.get("success"):
            raise InvalidApiToken()

        self.auth_token = "Token " + res["token"]
        return cast(str, res["customer_id"])

    async def get_email(self) -> str:
        """Return the user email"""
        if not self.auth_token:
            raise WebsocketError("token not set")

        res = await self._send_recv_single_message(
            {"command": "GET_ACCOUNT", "Authorization": self.auth_token}
        )

        if res["object"] == "ERROR":
            raise get_exception(res)

        if not res.get("login"):
            raise WebsocketError("No email found")
        return cast(str, res["login"])

    async def send_request(self, request: dict[str, Any]) -> None:
        """Add authorization and send request."""

        if not self.auth_token:
            raise WebsocketError("Token not set")

        await self.connected.wait()

        request["Authorization"] = self.auth_token
        await self._send(request)

    async def _message_handler(
        self,
        message: dict[str, Any],
        receiver: Callable[[dict[str, Any]], Coroutine[Any, Any, None]],
    ) -> None:
        """Handle message and pass to receiver."""

        object_name = message.get("object")

        if not object_name:
            raise WebsocketError("Received message has no object.")

        LOGGER.debug("Received %s", object_name)

        # handle ERROR object
        if object_name == "ERROR":
            raise get_exception(message)

        # if object other than ERROR has an error key it will be send to the receiver.
        error = message.get("error")

        # ignored objects
        if (
            ("RECEIVED" in object_name and not error)
            or object_name == "HELLO"
            or "OPERATIVE" in object_name
        ):
            return

        if object_name == "CHARGE_POINTS":
            handle_charge_points(message)
        elif object_name == "CH_STATUS":
            handle_status(message)
        elif object_name == "CH_SETTINGS":
            handle_settings(message)
        elif object_name == "CHARGE_CARDS":
            pass
        elif "GRID" in object_name:
            handle_grid(message)
        elif object_name in (
            "STATUS_SET_PUBLIC_CHARGING",
            "STATUS_SET_PLUG_AND_CHARGE",
        ):
            handle_setting_change(message)
        elif any(button in object_name for button in BUTTONS):
            handle_session_messages(message)
        else:
            return

        await receiver(message)

        if object_name == "CHARGE_POINTS":
            self.received_charge_points.set()

    async def _send(self, data: dict[str, Any]) -> None:
        """Send data to the websocket."""
        if self.conn:
            LOGGER.debug("Sending %s.", data["command"])
            data_str = json.dumps(data)
            await self.conn.send(data_str)
        else:
            raise WebsocketError("Connection is closed.")

    async def disconnect(self) -> None:
        """Disconnect from de websocket."""
        if not self.conn:
            raise WebsocketError("Connection is already closed.")
        await self.conn.close()

    def raise_correct_exception(self, err: Exception) -> None:
        """Check if the client was rejected by the server"""

        if isinstance(err, InvalidStatusCode):
            reason = err.headers.get("x-websocket-reject-reason")
            if reason is not None:
                if "Request limit reached" in reason:
                    raise RequestLimitReached("Request limit reached") from err
                if "Already connected" in reason:
                    raise AlreadyConnected("Already connected")
        if isinstance(err, ConnectionClosedError) and err.code == 4001:
            raise RequestLimitReached("Request limit reached") from err

        raise WebsocketError from err



"""Define package errors."""


class BlueCurrentException(Exception):
    """Define a base error."""


class WebsocketError(BlueCurrentException):
    """Define an error related to the websocket connection."""


class RequestLimitReached(BlueCurrentException):
    """Define an error for when the request limit is reached."""


class AlreadyConnected(BlueCurrentException):
    """Define an error for when the ip is already connected."""


class InvalidApiToken(BlueCurrentException):
    """Define an error related to an invalid token."""


class NoCardsFound(BlueCurrentException):
    """Define an error for when a token has no cards."""



from unittest.mock import MagicMock
from websockets.client import WebSocketClientProtocol
from websockets.legacy.client import Connect
from websockets.exceptions import (
    InvalidStatusCode,
    ConnectionClosedError,
    WebSocketException,
)
from websockets.frames import Close

from src.bluecurrent_api.websocket import (
    Websocket,
    WebsocketError,
    RequestLimitReached,
    InvalidApiToken,
    AlreadyConnected,
)

import pytest
from pytest_mock import MockerFixture


@pytest.mark.asyncio
async def test_start(mocker: MockerFixture):
    mocker.patch("src.bluecurrent_api.websocket.Websocket.validate_api_token")
    mock__loop = mocker.patch("src.bluecurrent_api.websocket.Websocket._loop")

    websocket = Websocket()

    mock_receiver = mocker.AsyncMock()
    mock_on_open = mocker.AsyncMock()

    with pytest.raises(WebsocketError):
        await websocket.start(mock_receiver, mock_on_open)

    websocket.auth_token = "123"
    await websocket.start(mock_receiver, mock_on_open)
    mock__loop.assert_called_once_with(mock_receiver, mock_on_open)

    mock_raise_correct_exception = mocker.patch(
        "src.bluecurrent_api.websocket.Websocket.raise_correct_exception"
    )
    err = WebSocketException()
    mock__loop.side_effect = err

    await websocket.start(mock_receiver, mock_on_open)
    mock_raise_correct_exception.assert_called_once_with(err)


@pytest.mark.asyncio
async def test__loop(mocker: MockerFixture):
    websocket = Websocket()
    mock_connect = MagicMock(spec=Connect)
    mocker.patch("src.bluecurrent_api.websocket.connect", return_value=mock_connect)
    mock_raise_correct_exception = mocker.patch(
        "src.bluecurrent_api.websocket.Websocket.raise_correct_exception"
    )

    mock_receiver = mocker.AsyncMock()
    mock_on_open = mocker.AsyncMock()

    mock_on_open.side_effect = WebSocketException()
    await websocket._loop(mock_receiver, mock_on_open)
    assert websocket.conn is None
    assert websocket.connected.is_set() is False
    assert websocket.received_charge_points.is_set() is False
    mock_raise_correct_exception.assert_called_once()


@pytest.mark.asyncio
async def test__send_recv_single_message(mocker: MockerFixture):
    websocket = Websocket()
    mock_connect = MagicMock(spec=Connect)
    mock_ws = MagicMock(spec=WebSocketClientProtocol)
    mocker.patch("src.bluecurrent_api.websocket.connect", return_value=mock_connect)
    mock_connect.__aenter__.return_value = mock_ws
    mock_ws.recv.return_value = '{"a": 1}'

    assert await websocket._send_recv_single_message({}) == {"a": 1}

    err = WebSocketException()
    mock_ws.recv.side_effect = err
    with pytest.raises(WebsocketError):
        await websocket._send_recv_single_message({})


@pytest.mark.asyncio
async def test_validate_token(mocker: MockerFixture):
    api_token = "123"
    websocket = Websocket()

    mocker.patch(
        "src.bluecurrent_api.websocket.Websocket._send_recv_single_message",
        return_value={
            "object": "STATUS_API_TOKEN",
            "success": True,
            "token": "abc",
            "customer_id": "123",
        },
    )
    result = await websocket.validate_api_token(api_token)
    assert result == "123"
    assert websocket.auth_token == "Token abc"

    mocker.patch(
        "src.bluecurrent_api.websocket.Websocket._send_recv_single_message",
        return_value={"object": "STATUS_API_TOKEN", "success": False, "error": ""},
    )
    with pytest.raises(InvalidApiToken):
        await websocket.validate_api_token(api_token)

    mocker.patch(
        "src.bluecurrent_api.websocket.Websocket._send_recv_single_message",
        return_value={
            "object": "ERROR",
            "error": 42,
            "message": "Request limit reached",
        },
    )
    with pytest.raises(RequestLimitReached):
        await websocket.validate_api_token(api_token)


@pytest.mark.asyncio
async def test_get_email(mocker: MockerFixture):
    websocket = Websocket()
    mocker.patch(
        "src.bluecurrent_api.websocket.Websocket._send_recv_single_message",
        return_value={"object": "ACCOUNT", "login": "test"},
    )

    with pytest.raises(WebsocketError):
        await websocket.get_email()
    websocket.auth_token = "abc"
    assert await websocket.get_email() == "test"

    mocker.patch(
        "src.bluecurrent_api.websocket.Websocket._send_recv_single_message",
        return_value={"object": "ACCOUNT"},
    )
    with pytest.raises(WebsocketError):
        await websocket.get_email()

    mocker.patch(
        "src.bluecurrent_api.websocket.Websocket._send_recv_single_message",
        return_value={
            "object": "ERROR",
            "error": 42,
            "message": "Request limit reached",
        },
    )
    with pytest.raises(RequestLimitReached):
        await websocket.get_email()


@pytest.mark.asyncio
async def test_send_request(mocker: MockerFixture):
    mock_send = mocker.patch("src.bluecurrent_api.websocket.Websocket._send")
    websocket = Websocket()
    websocket.connected.set()

    # without token
    with pytest.raises(WebsocketError):
        await websocket.send_request({"command": "GET_CHARGE_POINTS"})

    websocket.auth_token = "123"

    await websocket.send_request({"command": "GET_CHARGE_POINTS"})

    mock_send.assert_called_with(
        {"command": "GET_CHARGE_POINTS", "Authorization": "123"}
    )


@pytest.mark.asyncio
async def test_message_handler(mocker: MockerFixture):
    mock_handle_charge_points = mocker.patch(
        "src.bluecurrent_api.websocket.handle_charge_points"
    )
    mock_handle_status = mocker.patch("src.bluecurrent_api.websocket.handle_status")
    mock_handle_grid = mocker.patch("src.bluecurrent_api.websocket.handle_grid")
    mock_handle_settings = mocker.patch("src.bluecurrent_api.websocket.handle_settings")
    mock_handle_setting_change = mocker.patch(
        "src.bluecurrent_api.websocket.handle_setting_change"
    )

    mock_handle_handle_session_messages = mocker.patch(
        "src.bluecurrent_api.websocket.handle_session_messages"
    )

    websocket = Websocket()

    mock_receiver = mocker.AsyncMock()

    # CHARGE_POINTS flow
    message = {"object": "CHARGE_POINTS"}
    await websocket._message_handler(message, mock_receiver)
    mock_handle_charge_points.assert_called_with(message)
    mock_receiver.assert_called_with(message)

    message = {"object": "CHARGE_CARDS"}
    await websocket._message_handler(message, mock_receiver)
    mock_receiver.assert_called_with(message)

    # ch_status flow
    message = {"object": "CH_STATUS"}
    await websocket._message_handler(message, mock_receiver)
    mock_handle_status.assert_called_with(message)
    mock_receiver.assert_called_with(message)

    # grid_status flow
    message = {"object": "GRID_STATUS"}
    await websocket._message_handler(message, mock_receiver)
    mock_handle_grid.assert_called_with(message)
    mock_receiver.assert_called_with(message)

    # grid_current flow
    message = {"object": "GRID_CURRENT"}
    await websocket._message_handler(message, mock_receiver)
    mock_handle_grid.assert_called_with(message)
    mock_receiver.assert_called_with(message)

    # ch_settings flow
    message = {"object": "CH_SETTINGS"}
    await websocket._message_handler(message, mock_receiver)
    mock_handle_settings.assert_called_with(message)
    mock_receiver.assert_called_with(message)

    # setting change flow
    message = {"object": "STATUS_SET_PLUG_AND_CHARGE"}
    await websocket._message_handler(message, mock_receiver)
    mock_handle_setting_change.assert_called_with(message)
    mock_receiver.assert_called_with(message)

    # session message flow
    message = {"object": "STATUS_STOP_SESSION"}
    await websocket._message_handler(message, mock_receiver)
    mock_handle_handle_session_messages.assert_called_with(message)
    mock_receiver.assert_called_with(message)

    # STATUS_START_SESSION
    message = {"object": "STATUS_START_SESSION", "evse_id": "BCU101"}
    await websocket._message_handler(message, mock_receiver)
    mock_handle_handle_session_messages.assert_called_with(message)
    mock_receiver.assert_called_with(message)

    # no object
    message = {"value": True}
    with pytest.raises(WebsocketError):
        await websocket._message_handler(message, mock_receiver)

    # unknown command
    message = {"error": 0, "object": "ERROR", "message": "Unknown command"}
    with pytest.raises(WebsocketError):
        await websocket._message_handler(message, mock_receiver)

    # unknown token
    message = {"error": 1, "object": "ERROR", "message": "Invalid Auth Token"}
    with pytest.raises(WebsocketError):
        await websocket._message_handler(message, mock_receiver)

    # token not autorized
    message = {"error": 2, "object": "ERROR", "message": "Not authorized"}
    with pytest.raises(WebsocketError):
        await websocket._message_handler(message, mock_receiver)

    # unknown error
    message = {"error": 9, "object": "ERROR", "message": "Unknown error"}
    with pytest.raises(WebsocketError):
        await websocket._message_handler(message, mock_receiver)

    # limit reached
    message = {"error": 42, "object": "ERROR", "message": "Request limit reached"}
    with pytest.raises(RequestLimitReached):
        await websocket._message_handler(message, mock_receiver)

    # success false
    message = {"success": False, "error": "this is an error"}
    with pytest.raises(WebsocketError):
        await websocket._message_handler(message, mock_receiver)

    # Ignore status
    message = {"object": "STATUS"}
    await websocket._message_handler(message, mock_receiver)
    assert mock_receiver.call_count == 9

    # RECEIVED without error
    message = {"object": "RECEIVED_START_SESSION", "error": ""}
    await websocket._message_handler(message, mock_receiver)
    assert mock_receiver.call_count == 9


@pytest.mark.asyncio
async def test__send(mocker: MockerFixture):
    websocket = Websocket()

    with pytest.raises(WebsocketError):
        await websocket._send({"command": "test"})

    websocket.conn = mocker.MagicMock(spec=WebSocketClientProtocol)

    await websocket._send({"command": 1})
    websocket.conn.send.assert_called_with('{"command": 1}')


@pytest.mark.asyncio
async def test_disconnect(mocker: MockerFixture):
    websocket = Websocket()

    with pytest.raises(WebsocketError):
        await websocket.disconnect()

    websocket.conn = mocker.MagicMock(spec=WebSocketClientProtocol)
    await websocket.disconnect()
    websocket.conn.close.assert_called_once()


def test_raise_correct_exception():
    websocket = Websocket()

    with pytest.raises(RequestLimitReached):
        websocket.raise_correct_exception(
            InvalidStatusCode(
                403, {"x-websocket-reject-reason": "Request limit reached"}
            )
        )

    with pytest.raises(RequestLimitReached):
        websocket.raise_correct_exception(
            ConnectionClosedError(Close(4001, "Request limit reached"), None, None)
        )

    with pytest.raises(AlreadyConnected):
        websocket.raise_correct_exception(
            InvalidStatusCode(403, {"x-websocket-reject-reason": "Already connected"})
        )

    with pytest.raises(WebsocketError):
        websocket.raise_correct_exception(Exception)
